﻿/*
Frog Rescue - Game Controller class
*/

#include <windows.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <d3d9.h>
#pragma hdrstop

#include "tlib.h"
#include "LithGUI.h"
#include "LithDX.h"
#include "LithUtil.h"

#include "IStuff.h"

#include "CApplication.h"
#include "CGameController.h"
#include "CRoadMap.h"
#include "CFrog.h"
#include "CCar.h"

using namespace tlib;
using namespace LithDX;



CGameController::CGameController(CApplication* aOwner)
		: cDefaultCamZ(200), cSoundCounterMax(100 * 10)
{
	mOwner = aOwner;
	mFrogStage = 0;
	mFrogFrame = 0;

	const double Zn = 2, Zf = 1024;
	double a, b;
	b = 1.0 / (1.0 / Zn - 1.0 / Zf);
	a = b / Zn;
	// top-down (bottom-up?) projection matrix
	mProjectionMatrix(2, 2) = a;
	mProjectionMatrix(3, 2) = 1;
	mProjectionMatrix(0, 0) = 2;
	mProjectionMatrix(1, 1) = -2 * (1 / 0.75);
	mProjectionMatrix(2, 3) = -b;
	/*
	// first-person projection matrix
	mProjectionMatrix(2, 0) = a;
	mProjectionMatrix(3, 0) = 1;
	mProjectionMatrix(0, 1) = 2;
	mProjectionMatrix(1, 2) = 2 * (1 / 0.75);
	mProjectionMatrix(2, 3) = -b;
	*/

	// initialise game state
	mMouseX = 1024 / 2;
	mMouseY = 768 / 2;

	mCamX = 0;
	mCamY = 0;
	mCamZ = cDefaultCamZ;

	GenerateRoadData();

	mHasRoadBlock = false;

	mFrogsSaved = 0;
	mFrogsSplatted = 0;

	// get display/sound data
	mCursorSprite = mOwner->GetSprite(SL("Cursor"));

	mFrogTexture = mOwner->GetTexture(SL("Frog"));
	mJumpSound = mOwner->GetSound(SL("Jump"));

	mSoundCounter = cSoundCounterMax;

	for (int i = 0; i < 11; i++)
		mRoadTexture[i] = mOwner->GetTexture(TString::SPrintF(L"Road%d", i));

	for (int i = 0; i < 3; i++)
		mCarTexture[i] = mOwner->GetTexture(TString::SPrintF(L"Car%d", i));

	mSplatSound = mOwner->GetSound(SL("Splat"));

	mRoadBlockTexture = mOwner->GetTexture(SL("RoadBlock"));

	mFrogSavedSprite = mOwner->GetSprite(SL("FrogSaved"));
	mFrogSplattedSprite = mOwner->GetSprite(SL("FrogSplatted"));
}



void CGameController::GenerateRoadData()
{
	// road map
	mRoadMap = new CRoadMap(65, 65);
	mRoadMap->Generate(mOwner);

	// cars
	for (int i = 0; i < 8; i++)
	{
		int X, Y;
		SRoadNode Node = mRoadMap->GetRandomPos(mOwner, X, Y);
		int d = GetRandomDirection(Node);
		mCars.Add(CCar(-32 * 8 + X * 8, -32 * 8 + Y * 8, d * M_PI_2, 0, 20));
	}
	for (int i = 0; i < 6; i++)
	{
		int X, Y;
		SRoadNode Node = mRoadMap->GetRandomPos(mOwner, X, Y);
		int d = GetRandomDirection(Node);
		mCars.Add(CCar(-32 * 8 + X * 8, -32 * 8 + Y * 8, d * M_PI_2, 1, 16));
	}
	for (int i = 0; i < 2; i++)
	{
		int X, Y;
		SRoadNode Node = mRoadMap->GetRandomPos(mOwner, X, Y);
		int d = GetRandomDirection(Node);
		mCars.Add(CCar(-32 * 8 + X * 8, -32 * 8 + Y * 8, d * M_PI_2, 2, 7));
	}
}



/**
0 - Right
1 - Down
2 - Left
3 - Up
*/
int CGameController::GetRandomDirection(SRoadNode& aNode)
{
	int Max = 0;
	if (aNode.Right)
		Max++;
	if (aNode.Bottom)
		Max++;
	if (aNode.Left)
		Max++;
	if (aNode.Top)
		Max++;
	int d = mOwner->Random(0, Max - 1);
	if (!aNode.Right && d >= 0)
		d++;
	if (!aNode.Bottom && d >= 1)
		d++;
	if (!aNode.Left && d >= 2)
		d++;
	if (!aNode.Top && d >= 3)
		throw TException(L"Programmer stupidity error :(");
	return d;
}



void CGameController::Update()
{
	// scroll camera
	if (mMouseX < 50)
		mCamX -= 5 - mMouseX / 10;
	if (mMouseX > 1023 - 50)
		mCamX += (mMouseX - (1023 - 50)) / 10;
	if (mMouseY < 40)
		mCamY -= 5 - mMouseY / 8;
	if (mMouseY > 767 - 40)
		mCamY += (mMouseY - (767 - 40)) / 8;

	if (mCamZ > cDefaultCamZ)
	{
		mCamZ = TMax(mCamZ - mCamZ * 0.07, cDefaultCamZ);
	}

	for (int i = mFrogs.NumItems - 1; i >= 0; i--)
	{
		CFrog& Frog = mFrogs[i];
		if (Frog.GetX() < -35 * 8 || Frog.GetX() > 35 * 8 ||
				Frog.GetY() < -35 * 8 || Frog.GetY() > 35 * 8)
		{
			mFrogs.Remove(i);
			mFrogsSaved++;
			continue;
		}
		if (Frog.GetState() == fsIdle)
			Frog.Jump(mOwner->Random(20, 25), mOwner->Random(10, 15));
		Frog.Update();
	}
	mSoundCounter += mFrogs.NumItems;
	if (mSoundCounter >= cSoundCounterMax)
	{
		mSoundCounter = mSoundCounter % cSoundCounterMax;
		mSoundCounter -= mOwner->Random(0, 5) * 15;
		mOwner->PlaySound(mJumpSound);
	}

	if (mFrogStage < 5)
	{
		if (mFrogFrame != 0)
			mFrogFrame--;
		else
		{
			for (int i  = 0; i < 20; i++)
			{
				mFrogs.Add(CFrog(0, 0, 0, mOwner->Random() * 2 * M_PI));
				mFrogs.Last.Jump(mOwner->Random(20, 25), mOwner->Random(10, 15));
			}
			mFrogFrame = 40;
			mFrogStage++;
		}
	}

	// update cars
	for (int i = 0; i < mCars.NumItems; i++)
	{
		CCar& Car = mCars[i];
		if (Car.GetState() == csIdle)
		{
			int cx = (Car.GetX() + 260) / 8;
			int cy = (Car.GetY() + 260) / 8;
			SRoadNode Node = mRoadMap->GetData(cx, cy);
			// stop for roadblocks
			for (int i = 0; i < mRoadBlocks.NumItems; i++)
			{
				if (mRoadBlocks[i].X == cx - 1 && mRoadBlocks[i].Y == cy)
					Node.Left = false;
				if (mRoadBlocks[i].X == cx + 1 && mRoadBlocks[i].Y == cy)
					Node.Right = false;
				if (mRoadBlocks[i].X == cx && mRoadBlocks[i].Y == cy - 1)
					Node.Top = false;
				if (mRoadBlocks[i].X == cx && mRoadBlocks[i].Y == cy + 1)
					Node.Bottom = false;
			}
			int Count = 0;
			if (Node.Left)
				Count++;
			if (Node.Right)
				Count++;
			if (Node.Top)
				Count++;
			if (Node.Bottom)
				Count++;
			if (Count != 0)
			{
				int Dir;
				if (Count == 1)
					Dir = GetRandomDirection(Node);
				else
				{
					int CarDir = Car.GetAngle() / M_PI_2;
					if (CarDir == 0)
						Node.Left = false;
					if (CarDir == 1)
						Node.Top = false;
					if (CarDir == 2)
						Node.Right = false;
					if (CarDir == 3)
						Node.Bottom = false;
					Dir = GetRandomDirection(Node);
				}
				Car.SetAngle(Dir * M_PI_2);
				switch (Dir)
				{
				case 0:
					Car.Move(-32 * 8 + (cx + 1) * 8, -32 * 8 + cy * 8);
					break;
				case 1:
					Car.Move(-32 * 8 + cx * 8, -32 * 8 + (cy + 1) * 8);
					break;
				case 2:
					Car.Move(-32 * 8 + (cx - 1) * 8, -32 * 8 + cy * 8);
					break;
				case 3:
					Car.Move(-32 * 8 + cx * 8, -32 * 8 + (cy - 1) * 8);
					break;
				}
			}
		}
		Car.Update();
	}

	// frog-car collisions
	bool Splatted = false;
	for (int i = mFrogs.NumItems - 1; i >= 0; i--)
	{
		CFrog& Frog = mFrogs[i];
		for (int j = 0; j < mCars.NumItems; j++)
		{
			CCar& Car = mCars[j];
			bool CrossLeft = Frog.GetX() + 1 > Car.GetX() - 2.5;
			bool CrossRight = Frog.GetX() - 1 < Car.GetX() + 2.5;
			bool CrossTop = Frog.GetY() + 1 > Car.GetY() - 2.5;
			bool CrossBottom = Frog.GetY() - 1 < Car.GetY() + 2.5;
			if (CrossLeft && CrossRight && CrossTop && CrossBottom)
			{
				mFrogs.Remove(i);
				mFrogsSplatted++;
				if (!Splatted)
				{
					Splatted = true;
					mOwner->PlaySound(mSplatSound);
				}
			}
		}
	}

	// update roadblock stuff
	// what's this??? someone trying to fudge an inverse projection???
	double ScreenXSize = ((double)(mCamZ - 2) / (1024 - 2) * (512 - 1.5) + 1.5) * 2;
	double ScreenYSize = ((double)(mCamZ - 2) / (1024 - 2) * (384 - 1.125) + 1.125) * 2;
	// mouse in game-coordinates
	double MouseX = (double)(mMouseX - 512) / 1023 * ScreenXSize + mCamX;
	double MouseY = (double)(mMouseY - 384) / 767 * ScreenYSize + mCamY;
	int RoadX = (int)(MouseX + 260) / 8;
	int RoadY = (int)(MouseY + 260) / 8;
	if (!(RoadX >= 0 && RoadX < 65 && RoadY >= 0 && RoadY < 65))
		mHasRoadBlock = false;
	else
	{
		mRoadBlockX = RoadX;
		mRoadBlockY = RoadY;
		SRoadNode RBNode = mRoadMap->GetData(RoadX, RoadY);
		if (RBNode.Left && RBNode.Right && !RBNode.Top && !RBNode.Bottom)
		{
			mHasRoadBlock = true;
			mRoadBlockAngle = 1;
		}
		else if (!RBNode.Left && !RBNode.Right && RBNode.Top && RBNode.Bottom)
		{
			mHasRoadBlock = true;
			mRoadBlockAngle = 0;
		}
		else
			mHasRoadBlock = false;
	}

	if (mFrogs.NumItems == 0)
		mOwner->GameOver(mFrogsSaved, mFrogsSplatted);
}



void CGameController::Render(CDevice* aDevice)
{
	aDevice->SetTransform(D3DTS_PROJECTION, mProjectionMatrix);
	CMatrix ViewMatrix = CMatrix::Translate(-CVector(mCamX, mCamY, -mCamZ));
	aDevice->SetTransform(D3DTS_VIEW, ViewMatrix);

	// render grass
	aDevice->SetTransform(D3DTS_WORLD, CMatrix::Identity());
	CVertexColour V0(CVector(-300 * 8, -300 * 8, 0), D3DCOLOR_XRGB(84, 165, 67));
	CVertexColour V1(CVector(300 * 8, -300 * 8, 0), D3DCOLOR_XRGB(84, 165, 67));
	CVertexColour V2(CVector(-300 * 8, 300 * 8, 0), D3DCOLOR_XRGB(84, 165, 67));
	CVertexColour V3(CVector(300 * 8, 300 * 8, 0), D3DCOLOR_XRGB(84, 165, 67));
	aDevice->TriColour(V0, V1, V2);
	aDevice->TriColour(V1, V3, V2);

	// render roads
	for (int y = 0; y < mRoadMap->GetYSize(); y++)
	{
		for (int x = 0; x < mRoadMap->GetXSize(); x++)
		{
			SRoadNode Node = mRoadMap->GetData(x, y);
			if (!Node.Left && !Node.Right && !Node.Top && !Node.Bottom)
				continue;
			aDevice->SetTransform(D3DTS_WORLD, CMatrix::Translate(CVector(-32 * 8 + x * 8,
					-32 * 8 + y * 8, 0)));
			CVertexTexture V0(CVector(-4, -4, 0), 0, 0);
			CVertexTexture V1(CVector(4, -4, 0), 1, 0);
			CVertexTexture V2(CVector(-4, 4, 0), 0, 1);
			CVertexTexture V3(CVector(4, 4, 0), 1, 1);
			aDevice->TriTexture(V0, V1, V2, GetRoadTexture(Node));
			aDevice->TriTexture(V1, V3, V2, GetRoadTexture(Node));
		}
	}

	// render frogs
	for (int i = 0; i < mFrogs.NumItems; i++)
	{
		CFrog& Frog = mFrogs[i];
		aDevice->SetTransform(D3DTS_WORLD, CMatrix::RotateAround(CVector(0, 0, -1),
				Frog.GetAngle()) * CMatrix::Translate(CVector(Frog.GetX(), Frog.GetY(),
				-Frog.GetZ())));
		CVertexTexture V0(CVector(-1, -1, 0), 0, 0);
		CVertexTexture V1(CVector(1, -1, 0), 1, 0);
		CVertexTexture V2(CVector(-1, 1, 0), 0, 1);
		CVertexTexture V3(CVector(1, 1, 0), 1, 1);
		aDevice->TriTexture(V0, V1, V2, mFrogTexture);
		aDevice->TriTexture(V1, V3, V2, mFrogTexture);
	}

	// render cars
	for (int i = 0; i < mCars.NumItems; i++)
	{
		CCar& Car = mCars[i];
		aDevice->SetTransform(D3DTS_WORLD, CMatrix::RotateAround(CVector(0, 0, -1),
				Car.GetAngle()) * CMatrix::Translate(CVector(Car.GetX(), Car.GetY(), 0)));
		CVertexTexture V0(CVector(-2.5, -2.5, 0), 0, 0);
		CVertexTexture V1(CVector(2.5, -2.5, 0), 1, 0);
		CVertexTexture V2(CVector(-2.5, 2.5, 0), 0, 1);
		CVertexTexture V3(CVector(2.5, 2.5, 0), 1, 1);
		aDevice->TriTexture(V0, V1, V2, mCarTexture[Car.GetType()]);
		aDevice->TriTexture(V1, V3, V2, mCarTexture[Car.GetType()]);
	}

	for (int i = 0; i < mRoadBlocks.NumItems; i++)
	{
		SRoadBlock& RoadBlock = mRoadBlocks[i];
		aDevice->SetTransform(D3DTS_WORLD, CMatrix::RotateAround(CVector(0, 0, -1),
				RoadBlock.Angle * M_PI_2) * CMatrix::Translate(CVector(-32 * 8 +
				RoadBlock.X * 8, -32 * 8 + RoadBlock.Y * 8, 0)));
		CVertexTexture RV0(CVector(-2.3, -0.75, 0), 0, 0);
		CVertexTexture RV1(CVector(2.3, -0.75, 0), 1, 0);
		CVertexTexture RV2(CVector(-2.3, 0.75, 0), 0, 1);
		CVertexTexture RV3(CVector(2.3, 0.75, 0), 1, 1);
		aDevice->TriTexture(RV0, RV1, RV2, mRoadBlockTexture);
		aDevice->TriTexture(RV1, RV3, RV2, mRoadBlockTexture);
	}

	// render roadblock under cursor
	if (mHasRoadBlock)
	{
		aDevice->SetTransform(D3DTS_WORLD, CMatrix::RotateAround(CVector(0, 0, -1),
				mRoadBlockAngle * M_PI_2) * CMatrix::Translate(CVector(-32 * 8 +
				mRoadBlockX * 8, -32 * 8 + mRoadBlockY * 8, 0)));
		CVertexTexture RV0(CVector(-2.3, -0.75, 0), 0, 0);
		CVertexTexture RV1(CVector(2.3, -0.75, 0), 1, 0);
		CVertexTexture RV2(CVector(-2.3, 0.75, 0), 0, 1);
		CVertexTexture RV3(CVector(2.3, 0.75, 0), 1, 1);
		aDevice->TriTexture(RV0, RV1, RV2, mRoadBlockTexture);
		aDevice->TriTexture(RV1, RV3, RV2, mRoadBlockTexture);
	}

	// sprites and stuff
	aDevice->SetTransform(D3DTS_PROJECTION, CMatrix::Identity());
	aDevice->SetTransform(D3DTS_VIEW, CMatrix::Identity());
	aDevice->SetTransform(D3DTS_WORLD, CMatrix::Identity());

	aDevice->DrawSprite(mFrogSavedSprite, 850, 20, 0, 0);
	mOwner->DrawNumber(mFrogsSaved, 900, 24);
	aDevice->DrawSprite(mFrogSplattedSprite, 850, 70, 0, 0);
	mOwner->DrawNumber(mFrogsSplatted, 900, 74);

	aDevice->DrawSprite(mCursorSprite, mMouseX, mMouseY, 0.5, 0.5);
}



CTexture* CGameController::GetRoadTexture(SRoadNode& node)
{
	if (node.Top && node.Right && node.Bottom && node.Left)
		return mRoadTexture[10];

	if (node.Top && node.Right && node.Bottom)
		return mRoadTexture[6];
	if (node.Top && node.Right && node.Left)
		return mRoadTexture[7];
	if (node.Top && node.Bottom && node.Left)
		return mRoadTexture[8];
	if (node.Right && node.Bottom && node.Left)
		return mRoadTexture[9];

	if (node.Top && node.Right)
		return mRoadTexture[2];
	if (node.Top && node.Left)
		return mRoadTexture[3];
	if (node.Right && node.Bottom)
		return mRoadTexture[4];
	if (node.Bottom && node.Left)
		return mRoadTexture[5];

	if (node.Top || node.Bottom)
		return mRoadTexture[0];
	if (node.Left || node.Right)
		return mRoadTexture[1];

	throw TException(L"Unknown road texture");
}



void CGameController::SetMousePos(int aX, int aY)
{
	mMouseX = aX;
	if (mMouseX < 0)
		mMouseX = 0;
	if (mMouseX > 1023)
		mMouseX = 1023;
	mMouseY = aY;
	if (mMouseY < 0)
		mMouseY = 0;
	if (mMouseY > 767)
		mMouseY = 767;
}



void CGameController::CamGoesUp()
{
	mCamZ += 55;
}



void CGameController::PlaceRoadBlock()
{
	if (mHasRoadBlock)
	{
		for (int i = 0; i < mRoadBlocks.NumItems; i++)
		{
			if (mRoadBlockX == mRoadBlocks[i].X && mRoadBlockY == mRoadBlocks[i].Y)
			{
				mRoadBlocks.Remove(i);
				return;
			}
		}
		if (mRoadBlocks.NumItems >= 5)
			mRoadBlocks.Remove(0);
		mRoadBlocks.Add(SRoadBlock(mRoadBlockX, mRoadBlockY, mRoadBlockAngle));
	}
}