﻿/*
Frog Rescue - Roads
*/

#include <windows.h>
#pragma hdrstop

#include "tlib.h"

#include "IStuff.h"
#include "CRoadMap.h"

using namespace tlib;



CRoadMap::CRoadMap(int aXSize, int aYSize)
{
	if (!(aXSize % 2 == 1 && aYSize % 2 == 1))
		throw TException(L"Invalid roadmap size");
	mXSize = aXSize;
	mYSize = aYSize;
	mData = new SRoadNode [mXSize * mYSize];
	//memset(mData, 0, mXSize * mYSize * sizeof(SRoadNode));
}



CRoadMap::~CRoadMap()
{
	delete [] mData;
}



SRoadNode CRoadMap::GetData(int aX, int aY)
{
	if (!(aX >= 0 && aX < mXSize && aY >= 0 && aY < mYSize))
		throw TException(L"Invalid arguments");
	return mData[aX + aY * mXSize];
}



void CRoadMap::SetData(int aX, int aY, int aLeft, int aRight, int aTop, int aBottom)
{
	if (!(aX >= 0 && aX < mXSize && aY >= 0 && aY < mYSize))
		throw TException(L"Invalid arguments");
	if (!(aLeft >= -1 && aLeft <= 1 && aRight >= -1 && aRight <= 1 &&
			aTop >= -1 && aTop <= 1 && aBottom >= -1 && aBottom <= 1))
		throw TException(L"Invalid arguments");
	if (aLeft != -1)
	{
		mData[aX + aY * mXSize].Left = aLeft != 0;
		if (aX > 0)
			mData[(aX - 1) + aY * mXSize].Right = aLeft != 0;
	}
	if (aRight != -1)
	{
		mData[aX + aY * mXSize].Right = aRight != 0;
		if (aX < mXSize - 1)
			mData[(aX + 1) + aY * mXSize].Left = aRight != 0;
	}
	if (aTop != -1)
	{
		mData[aX + aY * mXSize].Top = aTop != 0;
		if (aY > 0)
			mData[aX + (aY - 1) * mXSize].Bottom = aTop != 0;
	}
	if (aBottom != -1)
	{
		mData[aX + aY * mXSize].Bottom = aBottom != 0;
		if (aY < mYSize - 1)
			mData[aX + (aY + 1) * mXSize].Top = aBottom != 0;
	}
}



SRoadNode CRoadMap::GetRandomPos(IStuff* aStuff, int& aX, int& aY)
{
	while (true)
	{
		int X = aStuff->Random(0, mXSize - 1);
		int Y;
		if (X % 2 == 0)
			Y = aStuff->Random(0, mYSize - 1);
		else
			Y = aStuff->Random(0, mYSize / 2) * 2;
		SRoadNode Node = GetData(X, Y);
		if (Node.Left || Node.Right || Node.Top || Node.Bottom)
		{
			aX = X;
			aY = Y;
			return Node;
		}
	}
}



void CRoadMap::Generate(IStuff* aStuff)
{
	AddRoadHorizontal(0, mXSize / 2, 0);
	AddRoadHorizontal(mYSize / 2, mXSize / 2, 0);
	AddRoadVertical(0, mYSize / 2, 0);
	AddRoadVertical(mXSize / 2, mYSize / 2, 0);

	for (int i = 0; i < 12; i++)
		CreateRandomRoad(aStuff, 24, 32);
	for (int i = 0; i < 20; i++)
		CreateRandomRoad(aStuff, 8, 12);
	for (int i = 0; i < 24; i++)
		CreateRandomRoad(aStuff, 1, 5);
}



void CRoadMap::CreateRandomRoad(IStuff* aStuff, int aMinLength, int aMaxLength)
{
	int RoadDir = aStuff->Random(0, 1);
	if (RoadDir == 0)
	{
		int NumColumns = mXSize / 2 + 1;
		int MaxLength = mXSize / 2;
		int Column = aStuff->Random(0, NumColumns - 1);
		int Length = aStuff->Random(aMinLength, aMaxLength);
		int Offset;
		if (Length == MaxLength)
			Offset = 0;
		else
			Offset = aStuff->Random(0, NumColumns - Length - 1);
		AddRoadVertical(Column, Length, Offset);
	}
	else
	{
		int NumRows = mYSize / 2 + 1;
		int MaxLength = mYSize / 2;
		int Row = aStuff->Random(0, NumRows - 1);
		int Length = aStuff->Random(aMinLength, aMaxLength);
		int Offset;
		if (Length == MaxLength)
			Offset = 0;
		else
			Offset = aStuff->Random(0, NumRows - Length - 1);
		AddRoadHorizontal(Row, Length, Offset);
		for (int j = 0; j < Length * 2 + 1; j++)
			SetData(Offset * 2 + j, Row * 2, j == 0 ? 0 : 1, j == Length * 2 ? 0 : 1, -1, -1);
	}
}



void CRoadMap::AddRoadVertical(int aColumn, int aLength, int aOffset)
{
	for (int j = 0; j <= aLength * 2; j++)
		SetData(aColumn * 2, aOffset * 2 + j, -1, -1, j == 0 ? -1 : 1, j == aLength * 2 ? -1 : 1);
}



void CRoadMap::AddRoadHorizontal(int aRow, int aLength, int aOffset)
{
	for (int j = 0; j < aLength * 2 + 1; j++)
		SetData(aOffset * 2 + j, aRow * 2, j == 0 ? 0 : 1, j == aLength * 2 ? 0 : 1, -1, -1);
}