/*
tlib - Unicode Version
5 February 2007
*/

#ifndef tlib__tlibH
#define tlib__tlibH

namespace tlib
{

typedef unsigned char TByte;



template <class T> inline T TAbs(const T& v)
{
	return v >= 0 ? v : -v;
}



template <class T> inline signed int TSign(const T& v)
{
	if (v > 0) return 1;
	if (v < 0) return -1;
	return 0;
}



template <class T> inline T TMin(const T& a, const T& b)
{
	return a <= b ? a : b;
}



template <class T> inline T TMax(const T& a, const T& b)
{
	return a >= b ? a : b;
}



template <class T> inline void TSwap(T& a, T& b)
{
	T t = a;
	a = b;
	b = t;
}



// language-independent properties

#if defined(__BORLANDC__)
	#define PROPERTY_F(name, type) __property type name = { read=_Get##name };
	#define PROPERTY_V(name, type, get_variable) __property type name = { read=get_variable };
	#define PROPERTY_C(name, type, get_variable) __property const type name = { read=_Get##name }; \
			private: const type _Get##name() const { return get_variable; } \
			public:
	#define PROPERTY_FF(name, type, set_type) \
			__property type name = { read=_Get##name, write=_Set##name };
	#define PROPERTY_VV(name, type, variable, set_type) \
			__property type name = { read=variable, write=variable };
	#define PROPERTY_VF(name, type, variable, set_type) \
			__property type name = { read=variable, write=_Set##name };
	//PROPERTY_FV(name, type, get_function, set_type, set_variable)
#elif defined(_MSC_VER)
	#define PROPERTY_F(name, type) __declspec(property(get=Get##name)) type name; \
			type Get##name() const { return _Get##name(); }
	#define PROPERTY_V(name, type, get_variable) __declspec(property(get=Get##name)) type name; \
			type Get##name() const { return get_variable; }
	#define PROPERTY_C(name, type, get_variable) __declspec(property(get=Get##name)) const type name; \
			const type Get##name() const { return get_variable; }
	#define PROPERTY_FF(name, type, set_type) \
			__declspec(property(get=Get##name, put=Set##name)) type name; \
			type Get##name() const { return _Get##name(); } \
			void Set##name(set_type value) { _Set##name(value); }
	#define PROPERTY_VV(name, type, variable, set_type) \
			__declspec(property(get=Get##name, put=Set##name)) type name; \
			type Get##name() const { return variable; } \
			void Set##name(set_type value) { variable = value; }
	#define PROPERTY_VF(name, type, variable, set_type) \
			__declspec(property(get=Get##name, put=Set##name)) type name; \
			type Get##name() const { return variable; } \
			void Set##name(set_type value) { _Set##name(value); }
	//PROPERTY_FV(name, type, get_function, set_type, set_variable)
#else
	#error Unknown compiler
#endif



class TException
{
public:
	TException();
	TException(const wchar_t* fmt, ...);
	TException(const TException& exception);
	~TException()
	{
		delete [] FMessage;
	}

	TException& operator =(const TException& exception);

	PROPERTY_C(Message, wchar_t*, FMessage)

private:
	wchar_t *FMessage;
};



class TString
{
public:
	TString();
	TString(int, const wchar_t*);
	TString(int, int=1);
	TString(const TString&);
	~TString();

	static TString SPrintF(const wchar_t*, ...);
	static TString IntToStr(int);

	TString& operator =(const TString&);
	TString operator +(const TString&) const;
	TString& operator +=(const TString&);

	bool operator ==(const TString&) const;
	bool operator !=(const TString&) const;

	int operator [](int) const;

	TString Substring(int) const;
	TString Substring(int, int) const;

	int FindSubstring(TString, int=0) const;
	int FindChar(int, int=0) const;
	int FindLastChar(int) const;

	PROPERTY_F(Str, const wchar_t*)

	PROPERTY_F(Length, int)
	PROPERTY_F(Empty, bool)
	PROPERTY_V(Size, int, FSize)



	int Iterator(int) const;
	bool HasNext(int i) const { return i < FSize; }
	void Next(int&) const;
	int Last() const;
	bool HasPrev(int i) const { return i >= 0; }
	void Prev(int&) const;
	int GetChar(int) const;

private:
	int *FReferenceCount;
	wchar_t *FData;
	int FSize;
	const wchar_t* _GetStr() const { return !FData ? L"" : FData; }

	int _GetLength() const;
	bool _GetEmpty() const { return FSize == 0; }

	void Release()
	{
		if (FReferenceCount)
		{
			if (*FReferenceCount == 1)
			{
				delete FReferenceCount;
				delete [] FData;
			}
			else
				--*FReferenceCount;
		}
	}

	int GetPos(int) const;
	int GetSize(int, int) const;
};



class TChar
{
public:
	static bool IsAlpha(int);
	static bool IsAlphaNumeric(int);
	static bool IsWhiteSpace(int);
};

}   // namespace tlib



#include "TSPtr.h"

// container classes

#include "TArray.h"
#include "TStack.h"
#include "TQueue.h"
#include "TSlowMap.h"
#include "TStringMap.h"
#include "THashTable.h"

#include "TStringBuilder.h"



#define SL(s) tlib::TString(sizeof(s) - 1, L##s)

#endif