/*
TStringBuilder Class
26 January 2008
*/
#include <stdlib.h>
#include <wchar.h>
#pragma hdrstop

#include "tlib.h"

using namespace tlib;



TStringBuilder::TStringBuilder()
{
	FData = NULL;
	FCapacity = 0;
	FSize = 0;
	FLength = 0;
}



TStringBuilder::~TStringBuilder()
{
	if (FData)
		delete [] FData;
}



void TStringBuilder::Clear()
{
	FSize = 0;
	FLength = 0;
}



TString TStringBuilder::GetString() const
{
	return TString(FSize, FData);
}



void TStringBuilder::AppendChar(int c)
{
	if (c >= 0xd800 && c < 0xe000)
		throw TException(L"Invalid character");
	if (c < 0x10000)
	{
		AllocData(FSize + 1);
		FData[FSize++] = c;
	}
	else
	{
		AllocData(FSize + 2);
		FData[FSize++] = 0xd800 | (c >> 10 & 0x3ff);
		FData[FSize++] = 0xdc00 | (c & 0x3ff);
	}
	FLength++;
}



void TStringBuilder::AppendChars(int c, int length)
{
	if (c >= 0xd800 && c < 0xe000)
		throw TException(L"Invalid character");
	if (length < 0)
		throw TException(L"Invalid length");
	if (c < 0x10000)
	{
		AllocData(FSize + length);
		for (int i = 0; i < length; i++)
			FData[FSize++] = c;
	}
	else
	{
		AllocData(FSize + length * 2);
		wchar_t e0 = 0xd800 | (c >> 10 & 0x3ff);
		wchar_t e1 = 0xdc00 | (c & 0x3ff);
		for (int i = 0; i < length; i++)
		{
			FData[FSize++] = e0;
			FData[FSize++] = e1;
		}
	}
	FLength += length;
}



void TStringBuilder::AppendData(int size, const wchar_t* data)
{
	AllocData(FSize + size);
	for (int i = 0; i < size;)
	{
		wchar_t e = data[i++];
		if (e >= 0xdc00 && e < 0xe000)
			throw TException(L"Invalid data");
		FData[FSize++] = e;
		if (e >= 0xd800 && e < 0xdc00)
		{
			if (i == size)
				throw TException(L"Invalid data");
			e = data[i++];
			if (!(e >= 0xdc00 && e < 0xe000))
				throw TException(L"Invalid data");
			FData[FSize++] = e;
		}
		FLength++;
	}
}



void TStringBuilder::AllocData(int minCapacity)
{
	if (FCapacity >= minCapacity)
		return;

	FCapacity = TMax(TMax(FCapacity + FCapacity, minCapacity), 8);

	wchar_t *OldData = FData;
	FData = new wchar_t [FCapacity];
	if (OldData)
	{
		wmemcpy(FData, OldData, FSize);
		delete [] OldData;
	}
}